# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import re
import subprocess
import tempfile
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import boto3
import sagemaker
import yaml
from botocore.exceptions import ClientError
from sagemaker.ai_registry.dataset import DataSet
from sagemaker.core.helper.session_helper import Session, get_execution_role
from sagemaker.core.shapes import (
    OutputDataConfig,
    S3DataSource,
    TensorBoardOutputConfig,
)
from sagemaker.core.training.configs import (
    Compute,
    InputData,
    Networking,
    StoppingCondition,
)
from sagemaker.train.model_trainer import ModelTrainer

from amzn_nova_forge.model.model_enums import Model, TrainingMethod
from amzn_nova_forge.recipe.recipe_config import HYPERPOD_RECIPE_PATH
from amzn_nova_forge.util.bedrock import (
    get_customization_type,
    parse_bedrock_recipe_config,
    resolve_base_model_identifier,
)
from amzn_nova_forge.util.logging import logger

# Maps TrainingMethod to (CustomizationTechnique, Peft|None) for ServerlessJobConfig
_METHOD_TO_SERVERLESS_CONFIG: Dict[TrainingMethod, tuple[str, Optional[str]]] = {
    TrainingMethod.SFT_LORA: ("SFT", "LORA"),
    TrainingMethod.SFT_FULL: ("SFT", None),
    TrainingMethod.DPO_LORA: ("DPO", "LORA"),
    TrainingMethod.DPO_FULL: ("DPO", None),
    # TrainingMethod.RFT_LORA: ("RLVR", "LORA"), # TODO: Add RLVR support
    # TrainingMethod.RFT_FULL: ("RLVR", None),
}
DEFAULT_SMTJ_JOB_MAX_RUNTIME = 86400  # 1 day


@dataclass
class JobConfig:
    job_name: str
    image_uri: str
    recipe_path: str
    output_s3_path: Optional[str] = None
    data_s3_path: Optional[str] = None
    input_s3_data_type: Optional[str] = None
    validation_data_s3_path: Optional[str] = (
        None  # Validation data S3 path (for CPT and Bedrock)
    )
    rft_lambda_arn: Optional[str] = None  # RFT Lambda ARN (for RFT jobs on Bedrock)
    mlflow_tracking_uri: Optional[str] = None  # MLflow tracking server ARN
    mlflow_experiment_name: Optional[str] = None
    mlflow_run_name: Optional[str] = None
    method: Optional[TrainingMethod] = None  # Training method (required for Bedrock)
    # TODO: The mlflow config is populated in recipe for both SMTJ and SMHP but will only work fro SMHP as SMTJ support for mlfow is only through boto3, fix this wit sagemaker 3 update


class RuntimeManager(ABC):
    def __init__(
        self,
        instance_type: Optional[str],
        instance_count: Optional[int],
        kms_key_id: Optional[str],
    ):
        self._instance_type = instance_type
        self._instance_count = instance_count
        self._kms_key_id = kms_key_id

    @property
    def instance_type(self) -> Optional[str]:
        """Type of instance (e.g., ml.p5.48xlarge)."""
        return self._instance_type

    @property
    def instance_count(self) -> Optional[int]:
        """Number of instances used."""
        return self._instance_count

    # Needed to update the instance_count if user decides to override its value
    @instance_count.setter
    def instance_count(self, value: Optional[int]) -> None:
        self._instance_count = value

    @property
    def kms_key_id(self) -> Optional[str]:
        """Optional KMS Key Id to use in S3 Bucket encryption, training jobs and deployments."""
        return self._kms_key_id

    @abstractmethod
    def setup(self) -> None:
        """Prepare environment and dependencies"""
        pass

    @abstractmethod
    def execute(self, job_config: JobConfig) -> str:
        """Launch a job and return a job id."""
        pass

    @abstractmethod
    def cleanup(self, job_id: str) -> None:
        """Tear down or release resources."""
        pass

    @classmethod
    def _s3_bucket_arn_from_path(cls, s3_path):
        """Extract S3 bucket ARN from a single S3 path."""
        if not s3_path:
            return None
        bucket = s3_path.split("/")[2]
        return f"arn:aws:s3:::{bucket}"

    @classmethod
    def _s3_object_arn_from_path(cls, s3_path):
        """Extract S3 object ARN from a single S3 path."""
        if not s3_path:
            return None
        bucket = s3_path.split("/")[2]
        # Allow access to the specific path and subdirectories
        if len(s3_path.split("/")) > 3:
            # Has a path component, use it
            path = "/".join(s3_path.split("/")[3:])
            return f"arn:aws:s3:::{bucket}/{path}*"
        else:
            # Just bucket, allow all objects
            return f"arn:aws:s3:::{bucket}/*"

    @classmethod
    def required_calling_role_permissions(cls, data_s3_path=None, output_s3_path=None):
        """Base permissions required by all runtime managers."""
        permissions = []

        # Collect unique bucket ARNs
        bucket_arns = set()
        for s3_path in [data_s3_path, output_s3_path]:
            bucket_arn = cls._s3_bucket_arn_from_path(s3_path)
            if bucket_arn:
                bucket_arns.add(bucket_arn)

        # Add bucket-level permissions
        for bucket_arn in bucket_arns:
            permissions.extend(
                [
                    ("s3:CreateBucket", bucket_arn),
                    ("s3:ListBucket", bucket_arn),
                ]
            )

        # Add input-specific permissions (read-only)
        if data_s3_path:
            data_object_arn = cls._s3_object_arn_from_path(data_s3_path)
            permissions.append(("s3:GetObject", data_object_arn))

        # Add output-specific permissions (read-write)
        if output_s3_path:
            output_object_arn = cls._s3_object_arn_from_path(output_s3_path)
            permissions.extend(
                [
                    ("s3:GetObject", output_object_arn),
                    ("s3:PutObject", output_object_arn),
                ]
            )

        return permissions


class SMTJRuntimeManager(RuntimeManager):
    def __init__(
        self,
        instance_type: str,
        instance_count: int,
        execution_role: Optional[str] = None,
        kms_key_id: Optional[str] = None,
        encrypt_inter_container_traffic: bool = False,
        subnets: Optional[list[str]] = None,
        security_group_ids: Optional[list[str]] = None,
        max_job_runtime: Optional[int] = DEFAULT_SMTJ_JOB_MAX_RUNTIME,
    ):
        # NOTE: Not setting execution_role directly due to issues with mypy type inference
        self._execution_role = execution_role

        self.subnets = subnets
        self.security_group_ids = security_group_ids
        self.encrypt_inter_container_traffic = encrypt_inter_container_traffic
        self.max_job_runtime = max_job_runtime

        super().__init__(instance_type, instance_count, kms_key_id)
        self.setup()

    @classmethod
    def required_calling_role_permissions(cls, data_s3_path=None, output_s3_path=None):
        """Required permissions for SMTJ calling role operations and execution role validation."""
        # Start with base S3 permissions
        permissions = super().required_calling_role_permissions(
            data_s3_path, output_s3_path
        )

        # Add SMTJ-specific permissions
        permissions.extend(
            [
                ("sagemaker:CreateTrainingJob", "*"),
                ("sagemaker:DescribeTrainingJob", "*"),
                "iam:GetRole",
                "iam:PassRole",
                "iam:GetPolicy",
                "iam:GetPolicyVersion",
                "iam:ListRolePolicies",
                "iam:GetRolePolicy",
                "iam:ListAttachedRolePolicies",
            ]
        )

        return permissions

    def setup(self) -> None:
        boto_session = boto3.session.Session()
        self.region = boto_session.region_name or "us-east-1"
        self.sagemaker_client = boto3.client("sagemaker", region_name=self.region)
        self.sagemaker_session = Session(
            boto_session=boto_session, sagemaker_client=self.sagemaker_client
        )

        if self._execution_role is None:
            self.execution_role = get_execution_role(use_default=True)
        else:
            self.execution_role = self._execution_role
        # Delete temporary attribute so customers don't confuse it with the actual attribute
        del self._execution_role

    def execute(self, job_config: JobConfig) -> str:
        from amzn_nova_forge.validation.validator import Validator

        Validator.validate_job_name(job_name=job_config.job_name)

        try:
            assert job_config.output_s3_path is not None

            tensorboard_output_config = TensorBoardOutputConfig(
                s3_output_path=job_config.output_s3_path,
            )

            compute = Compute(
                instance_count=self.instance_count,
                instance_type=self.instance_type,
            )
            output_data_config = OutputDataConfig(
                s3_output_path=job_config.output_s3_path,
                kms_key_id=self.kms_key_id,
            )
            networking = Networking(
                subnets=self.subnets,
                security_group_ids=self.security_group_ids,
            )
            stopping_condition = StoppingCondition(
                max_runtime_in_seconds=self.max_job_runtime
            )

            trainer_config = {
                "training_recipe": job_config.recipe_path,
                "compute": compute,
                "networking": networking,
                "stopping_condition": stopping_condition,
                "output_data_config": output_data_config,
                "base_job_name": job_config.job_name,
                "role": self.execution_role,
                "sagemaker_session": self.sagemaker_session,
                "training_image": job_config.image_uri,
            }
            # For eval job, the input could be none
            # https://docs.aws.amazon.com/sagemaker/latest/dg/nova-model-evaluation.html#nova-model-evaluation-notebook
            if job_config.data_s3_path:
                input_data = InputData(
                    channel_name="train",
                    data_source=S3DataSource(
                        s3_uri=job_config.data_s3_path,
                        s3_data_type=job_config.input_s3_data_type,
                        s3_data_distribution_type="FullyReplicated",
                    ),
                )
                trainer_config["input_data_config"] = [input_data]

            model_trainer = ModelTrainer.from_recipe(
                **trainer_config
            ).with_tensorboard_output_config(tensorboard_output_config)
            model_trainer.train(wait=False, logs=False)
            # Added since the job name has an auto generated suffix by sagemaker v3
            sagemaker_client = boto3.client("sagemaker")
            list_jobs_response = sagemaker_client.list_training_jobs(
                NameContains=job_config.job_name,
                SortBy="CreationTime",
                SortOrder="Descending",
                MaxResults=1,
            )
            unique_job_name = list_jobs_response["TrainingJobSummaries"][0][
                "TrainingJobName"
            ]
            return unique_job_name

        except Exception as e:
            logger.error(f"Failed to start training job: {str(e)}")
            raise

    def cleanup(self, job_name: str) -> None:
        try:
            self.sagemaker_client.stop_training_job(TrainingJobName=job_name)
            self.sagemaker_client.close()
        except Exception as e:
            logger.error(f"Failed to cleanup job {job_name}: {str(e)}")
            raise


# TODO: Might need to take RIG as input in case of multiple RIGs
class SMHPRuntimeManager(RuntimeManager):
    def __init__(
        self,
        instance_type: str,
        instance_count: int,
        cluster_name: str,
        namespace: str,
        kms_key_id: Optional[str] = None,
    ):
        from amzn_nova_forge.validation.validator import Validator

        Validator.validate_cluster_name(cluster_name=cluster_name)
        Validator.validate_namespace(namespace=namespace)

        self.cluster_name = cluster_name
        self.namespace = namespace
        super().__init__(instance_type, instance_count, kms_key_id)
        self.setup()

    @classmethod
    def required_calling_role_permissions(cls, data_s3_path=None, output_s3_path=None):
        """Required permissions for HyperPod operations."""
        # Start with base S3 permissions
        permissions = super().required_calling_role_permissions(
            data_s3_path, output_s3_path
        )

        # Add SMHP-specific permissions
        permissions.extend(
            [
                (
                    "sagemaker:DescribeCluster",
                    lambda infra: (
                        f"arn:aws:sagemaker:{infra.region}:*:cluster/{infra.cluster_name}"
                    ),
                ),
                (
                    "eks:DescribeCluster",
                    lambda infra: f"arn:aws:eks:{infra.region}:*:cluster/*",
                ),
                (
                    "eks:ListAddons",
                    lambda infra: (
                        f"arn:aws:eks:{infra.region}:*:cluster/{infra.cluster_name}"
                    ),
                ),
                ("sagemaker:ListClusters", "*"),
            ]
        )

        return permissions

    def setup(self) -> None:
        boto_session = boto3.session.Session()
        self.region = boto_session.region_name or "us-east-1"

        response = subprocess.run(
            [
                "hyperpod",
                "connect-cluster",
                "--cluster-name",
                self.cluster_name,
                "--namespace",
                self.namespace,
            ],
            capture_output=True,
            text=True,
            check=True,
        )

        if response.stderr:
            logger.error(
                f"Unable to connect to HyperPod cluster {self.cluster_name}: {response.stderr}"
            )
            raise RuntimeError(response.stderr)

        logger.info(
            f"Successfully connected to HyperPod cluster '{self.cluster_name}' in namespace '{self.namespace}'."
        )

    def execute(self, job_config: JobConfig) -> str:
        try:
            # Scrub recipe path so that it will be recognized by the HyperPod CLI
            recipe_path = (
                job_config.recipe_path.split(HYPERPOD_RECIPE_PATH, 1)[1]
                .lstrip("/")
                .lstrip("\\")
                .removesuffix(".yaml")
            )

            override_parameters = json.dumps(
                {
                    "instance_type": self.instance_type,
                    "container": job_config.image_uri,
                }
            )
            response = subprocess.run(
                [
                    "hyperpod",
                    "start-job",
                    "--namespace",
                    self.namespace,
                    "--recipe",
                    recipe_path,
                    "--override-parameters",
                    override_parameters,
                ],
                capture_output=True,
                text=True,
                check=True,
            )

            if matched_job_name := re.search(r"NAME: (\S+)", response.stdout):
                return matched_job_name.group(1)
            raise ValueError(
                f"Could not find job name in output. There may be an issue with the helm installation, "
                f"assumed role permissions to trigger jobs on the cluster, or job specification. Output: {response.stdout}"
            )
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to start HyperPod job: {e.stderr}")
            raise

    def cleanup(self, job_name: str) -> None:
        from amzn_nova_forge.validation.validator import Validator

        Validator.validate_job_name(job_name=job_name)

        try:
            response = subprocess.run(
                [
                    "hyperpod",
                    "cancel-job",
                    "--job-name",
                    job_name,
                    "--namespace",
                    self.namespace,
                ],
                capture_output=True,
                text=True,
                check=True,
            )

            if response.stderr:
                logger.error(f"Failed to cleanup HyperPod job: {response.stderr}")

        except Exception as e:
            logger.error(f"Failed to cleanup HyperPod job '{job_name}': {str(e)}")
            raise


class BedrockRuntimeManager(RuntimeManager):
    """
    Runtime manager for AWS Bedrock model customization jobs.
    """

    def __init__(
        self,
        execution_role: str,
        base_model_identifier: Optional[str] = None,
        kms_key_id: Optional[str] = None,
        vpc_config: Optional[Dict[str, list[str]]] = None,
    ):
        # Store Bedrock-specific configuration
        self.execution_role = execution_role
        self.base_model_identifier = base_model_identifier
        self.vpc_config = vpc_config

        # Calls constructor with None for instance_type and instance_count
        # since Bedrock manages compute resources automatically
        super().__init__(instance_type=None, instance_count=None, kms_key_id=kms_key_id)

        self.setup()

    def setup(self) -> None:
        """Initialize Bedrock client and session.

        Creates a boto3 session, extracts the region, and initializes the Bedrock client.
        Falls back to us-east-1 if no region is configured.

        Raises:
            Exception: If Bedrock client initialization fails
        """
        try:
            boto_session = boto3.session.Session()
            self.region = boto_session.region_name or "us-east-1"
            self.bedrock_client = boto3.client("bedrock", region_name=self.region)
            logger.info(
                f"Successfully initialized Bedrock client in region {self.region}"
            )
        except Exception as e:
            logger.error(f"Failed to initialize Bedrock client: {e}")
            raise

    def execute(self, job_config: JobConfig) -> str:

        try:
            # Validate job name
            from amzn_nova_forge.validation.validator import Validator

            Validator.validate_job_name(job_name=job_config.job_name)
            logger.info(f"Starting Bedrock customization job: {job_config.job_name}")

            # Get training method from job_config (passed via nova_model_customizer)
            # For Bedrock, we need the method to determine customization type
            if job_config.method is None:
                raise ValueError(
                    "Training method must be provided in job_config for Bedrock jobs. "
                    "This should be set automatically by NovaModelCustomizer."
                )
            method = job_config.method

            # Extract hyperparameters from recipe if provided
            hyperparameters: Dict[str, str] = {}
            rft_hyperparameters: Dict[
                str, Any
            ] = {}  # RFT hyperparameters use native types (int, float, str)
            if job_config.recipe_path is not None:
                recipe_data = parse_bedrock_recipe_config(
                    job_config.recipe_path, method
                )
                hyperparameters = recipe_data["hyperparameters"]
                rft_hyperparameters = recipe_data["rft_hyperparameters"]
            else:
                logger.info(
                    "No recipe provided - Bedrock will use default hyperparameters"
                )

            # Get customization type from training method
            customization_type = get_customization_type(method)

            # Resolve base model identifier
            if job_config.recipe_path is not None:
                base_model_identifier = resolve_base_model_identifier(
                    job_config.recipe_path, self.base_model_identifier, self.region
                )
            elif self.base_model_identifier is not None:
                base_model_identifier = self.base_model_identifier
            else:
                raise ValueError(
                    "base_model_identifier must be provided either in BedrockRuntimeManager "
                    "constructor or via recipe_path. Cannot determine which model to use."
                )

            # Use job_name as custom_model_name
            custom_model_name = job_config.job_name

            # Build API request dictionary with required fields
            api_request: Dict[str, Any] = {
                "jobName": job_config.job_name,
                "customModelName": custom_model_name,
                "roleArn": self.execution_role,
                "baseModelIdentifier": base_model_identifier,
                "customizationType": customization_type,
            }

            # Add hyperParameters only if provided (optional parameter)
            if hyperparameters:
                api_request["hyperParameters"] = hyperparameters

            # Add trainingDataConfig (REQUIRED by Bedrock API)
            if job_config.data_s3_path is None:
                raise ValueError(
                    "data_s3_path is required for Bedrock customization jobs. "
                    "Bedrock API requires trainingDataConfig.s3Uri to be provided."
                )

            training_data_config: Dict[str, str] = {"s3Uri": job_config.data_s3_path}
            api_request["trainingDataConfig"] = training_data_config

            # Add outputDataConfig with output_s3_path (required)
            if job_config.output_s3_path is None:
                raise ValueError(
                    "output_s3_path is required for Bedrock customization jobs"
                )

            output_data_config: Dict[str, str] = {"s3Uri": job_config.output_s3_path}

            # Add KMS key to outputDataConfig only if provided (optional)
            if self.kms_key_id is not None:
                output_data_config["kmsKeyId"] = self.kms_key_id

            api_request["outputDataConfig"] = output_data_config

            # Add validationDataConfig only if validation_data_s3_path is provided (optional)
            if job_config.validation_data_s3_path is not None:
                # Warn that Nova Lite 2 doesn't support validation datasets
                if (
                    "nova-2-lite" in base_model_identifier
                    or "nova-lite-2" in base_model_identifier
                ):
                    logger.warning(
                        "Validation datasets are not supported for Nova Lite 2 models on Bedrock. "
                        "The validation_data_s3_path parameter will be ignored. "
                        "To use validation datasets, please use a different model (Nova Micro, Nova Lite, or Nova Pro)."
                    )
                else:
                    validation_data_config: Dict[str, List[Dict[str, str]]] = {
                        "validators": [{"s3Uri": job_config.validation_data_s3_path}]
                    }
                    api_request["validationDataConfig"] = validation_data_config

            # Add vpcConfig only if provided (optional)
            if self.vpc_config is not None:
                vpc_config: Dict[str, List[str]] = {}
                if "subnet_ids" in self.vpc_config:
                    vpc_config["subnetIds"] = self.vpc_config["subnet_ids"]
                if "security_group_ids" in self.vpc_config:
                    vpc_config["securityGroupIds"] = self.vpc_config[
                        "security_group_ids"
                    ]

                if vpc_config:  # Only add if we have at least one field
                    api_request["vpcConfig"] = vpc_config

            # Add customizationConfig for RFT jobs only if Lambda ARN provided (optional but required for RFT)
            if customization_type == "REINFORCEMENT_FINE_TUNING":
                if not job_config.rft_lambda_arn:
                    raise ValueError(
                        "rft_lambda_arn is required for RFT (Reinforcement Fine-Tuning) jobs. "
                        "Please provide the Lambda ARN in train() method."
                    )

                rft_config: Dict[str, Any] = {
                    "graderConfig": {
                        "lambdaGrader": {"lambdaArn": job_config.rft_lambda_arn}
                    }
                }

                # Add RFT hyperparameters if provided (native types: int, float, string)
                if rft_hyperparameters:
                    rft_config["hyperParameters"] = rft_hyperparameters

                customization_config: Dict[str, Any] = {"rftConfig": rft_config}
                api_request["customizationConfig"] = customization_config

            # Call Bedrock API to create customization job
            response = self.bedrock_client.create_model_customization_job(**api_request)

            # Extract and return job ARN
            job_arn = response["jobArn"]

            return job_arn

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]
            logger.error(f"Bedrock API error ({error_code}): {error_message}")

            if error_code == "ValidationException":
                raise ValueError(
                    f"Invalid Bedrock job configuration: {error_message}"
                ) from e
            elif error_code == "AccessDeniedException":
                raise PermissionError(f"Bedrock access denied: {error_message}") from e
            elif error_code == "ResourceNotFoundException":
                raise ValueError(f"Bedrock resource not found: {error_message}") from e
            elif error_code == "ServiceQuotaExceededException":
                raise RuntimeError(
                    f"Bedrock service quota exceeded: {error_message}"
                ) from e
            elif error_code == "ThrottlingException":
                raise RuntimeError(f"Bedrock API throttled: {error_message}") from e
            else:
                raise RuntimeError(f"Bedrock API error: {error_message}") from e

        except ValueError as e:
            # Re-raise ValueError (from validation, unsupported methods, etc.)
            logger.error(f"Validation error: {e}")
            raise

        except FileNotFoundError as e:
            # Re-raise FileNotFoundError (from recipe parsing)
            logger.error(f"Recipe file not found: {e}")
            raise

        except Exception as e:
            # Catch-all for unexpected errors
            logger.error(f"Unexpected error creating Bedrock customization job: {e}")
            raise RuntimeError(
                f"Failed to create Bedrock customization job: {e}"
            ) from e

    def cleanup(self, job_id: str) -> None:

        try:
            logger.info(f"Stopping Bedrock customization job: {job_id}")

            # Validate job_id format (should be a Bedrock job ARN)
            if not job_id or not isinstance(job_id, str):
                raise ValueError(
                    f"Invalid job_id: must be a non-empty string, got {type(job_id)}"
                )

            if not job_id.startswith("arn:aws:bedrock:"):
                logger.warning(
                    f"job_id does not appear to be a valid Bedrock ARN: {job_id}. "
                    "Expected format: arn:aws:bedrock:region:account:model-customization-job/job-id"
                )

            # Call Bedrock API to stop the customization job
            self.bedrock_client.stop_model_customization_job(jobIdentifier=job_id)
            logger.info(f"Successfully stopped Bedrock customization job: {job_id}")

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            error_message = e.response["Error"]["Message"]
            logger.error(
                f"Bedrock API error while stopping job ({error_code}): {error_message}"
            )

            if error_code == "ValidationException":
                raise ValueError(f"Invalid job ARN: {error_message}") from e
            elif error_code == "ResourceNotFoundException":
                raise ValueError(f"Job not found: {error_message}") from e
            elif error_code == "AccessDeniedException":
                raise PermissionError(
                    f"Insufficient permissions to stop job: {error_message}"
                ) from e
            elif error_code == "ConflictException":
                # Job may already be stopped or in a terminal state
                logger.warning(f"Job may already be stopped: {error_message}")
            else:
                raise RuntimeError(
                    f"Failed to stop Bedrock job: {error_message}"
                ) from e

        except Exception as e:
            logger.error(f"Unexpected error stopping Bedrock job: {e}")
            raise RuntimeError(f"Failed to stop Bedrock customization job: {e}") from e

        finally:
            # Close bedrock_client connection
            try:
                if hasattr(self, "bedrock_client") and self.bedrock_client is not None:
                    self.bedrock_client.close()
                    logger.info("Closed Bedrock client connection")
            except Exception as e:
                logger.warning(f"Error closing Bedrock client: {e}")

    @classmethod
    def required_calling_role_permissions(cls, data_s3_path=None, output_s3_path=None):
        """Return required IAM permissions for Bedrock operations."""
        # Start with base S3 permissions
        permissions = super().required_calling_role_permissions(
            data_s3_path, output_s3_path
        )

        # Add Bedrock-specific permissions
        permissions.extend(
            [
                ("bedrock:CreateModelCustomizationJob", "*"),
                ("bedrock:StopModelCustomizationJob", "*"),
                ("bedrock:GetModelCustomizationJob", "*"),
                "iam:PassRole",
            ]
        )

        return permissions


def _get_hub_content(
    hub_name: str,
    hub_content_name: str,
    hub_content_type: str,
    region: str,
) -> Dict[str, Any]:
    """
     Get hub content from SageMaker via the DescribeHubContent API

    Args:
        hub_name: Name of the SageMaker Hub
        hub_content_name: Name of the hub content
        hub_content_type: Type of hub content
        region: AWS region

    Returns:
        Dict containing hub content
    """
    sagemaker_client = boto3.client("sagemaker", region_name=region)

    try:
        response = sagemaker_client.describe_hub_content(
            HubName=hub_name,
            HubContentType=hub_content_type,
            HubContentName=hub_content_name,
        )

        # Parse HubContentDocument if it's a JSON string
        if "HubContentDocument" in response:
            hub_content_document = response["HubContentDocument"]
            if isinstance(hub_content_document, str):
                try:
                    response["HubContentDocument"] = json.loads(hub_content_document)
                except (json.JSONDecodeError, TypeError):
                    # If parsing fails, leave the string as is
                    pass

    except Exception as e:
        raise RuntimeError(
            f"Failed to get SageMaker hub content for '{hub_content_name}': {str(e)}"
        )

    return response


class SMTJServerlessRuntimeManager(RuntimeManager):
    def __init__(
        self,
        model_package_group_name: str,
        execution_role: Optional[str] = None,
        kms_key_id: Optional[str] = None,
        encrypt_inter_container_traffic: bool = False,
        subnets: Optional[list[str]] = None,
        security_group_ids: Optional[list[str]] = None,
        max_job_runtime: Optional[int] = DEFAULT_SMTJ_JOB_MAX_RUNTIME,  # 1 day
    ):
        # NOTE: Not setting execution_role directly due to issues with mypy type inference
        self._execution_role = execution_role
        self.model_package_group_name = model_package_group_name
        self.subnets = subnets
        self.security_group_ids = security_group_ids
        self.encrypt_inter_container_traffic = encrypt_inter_container_traffic
        self.max_job_runtime = max_job_runtime

        super().__init__(None, None, kms_key_id=kms_key_id)
        self.setup()

    @classmethod
    def required_calling_role_permissions(cls, data_s3_path=None, output_s3_path=None):
        """Required permissions for SMTJ calling role operations and execution role validation."""
        # Start with base S3 permissions
        permissions = super().required_calling_role_permissions(
            data_s3_path, output_s3_path
        )

        # Add SMTJ-specific permissions
        permissions.extend(
            [
                ("sagemaker:CreateTrainingJob", "*"),
                ("sagemaker:DescribeTrainingJob", "*"),
                "iam:GetRole",
                "iam:PassRole",
                "iam:GetPolicy",
                "iam:GetPolicyVersion",
                "iam:ListRolePolicies",
                "iam:GetRolePolicy",
                "iam:ListAttachedRolePolicies",
            ]
        )

        return permissions

    def setup(self) -> None:
        boto_session = boto3.session.Session()
        self.region = boto_session.region_name or "us-east-1"
        self.sagemaker_client = boto3.client("sagemaker", region_name=self.region)
        self.sagemaker_session = Session(
            boto_session=boto_session, sagemaker_client=self.sagemaker_client
        )

        if self._execution_role is None:
            self.execution_role = get_execution_role(use_default=True)
        else:
            self.execution_role = self._execution_role
        # Delete temporary attribute so customers don't confuse it with the actual attribute
        del self._execution_role

        # Create model package group if it doesn't exist
        try:
            resp = self.sagemaker_client.describe_model_package_group(
                ModelPackageGroupName=self.model_package_group_name
            )
            self.model_package_group_arn = resp["ModelPackageGroupArn"]
        except self.sagemaker_client.exceptions.ClientError:
            resp = self.sagemaker_client.create_model_package_group(
                ModelPackageGroupName=self.model_package_group_name
            )
            self.model_package_group_arn = resp["ModelPackageGroupArn"]

    def _resolve_base_model_arn(self, model: Model) -> str:
        """Resolve the BaseModelArn from SageMaker Hub for the given model."""
        hub_content = _get_hub_content(
            hub_name="SageMakerPublicHub",
            hub_content_name=model.hub_content_name,
            hub_content_type="Model",
            region=self.region,
        )
        return hub_content["HubContentArn"]

    def _extract_hyperparameters(self, recipe: Dict[str, Any]) -> Dict[str, str]:
        """Extract hyperparameters from recipe as string key-value pairs, checking two levels of keys."""
        # Recipe keys that map to create_training_job HyperParameters
        hyperparameter_map = {
            "global_batch_size": "global_batch_size",
            "lr": "learning_rate",
            "alpha": "lora_alpha",
            "max_length": "max_context_length",
            "max_steps": "max_steps",
            "name": "name",
            "reasoning_enabled": "reasoning_enabled",
            "max_epochs": "max_epochs",
            "warmup_steps": "warmup_steps",
            "lora_plus_lr_ratio": "lora_plus_lr_ratio",
        }
        result: Dict[str, str] = {}
        # Check up to 3 levels deep for hyperparameter keys
        for k, v in recipe.items():
            if k in hyperparameter_map.keys() and v is not None:
                result[hyperparameter_map[k]] = str(v)
            elif isinstance(v, dict):
                for k2, v2 in v.items():
                    if k2 in hyperparameter_map.keys() and v2 is not None:
                        result[hyperparameter_map[k2]] = str(v2)
                    elif isinstance(v2, dict):
                        for k3, v3 in v2.items():
                            if k3 in hyperparameter_map.keys() and v3 is not None:
                                result[hyperparameter_map[k3]] = str(v3)
        return result

    def _build_serverless_job_config(
        self, method: TrainingMethod, base_model_arn: str
    ) -> Dict[str, Any]:
        """Build ServerlessJobConfig for create_training_job."""
        technique, peft = _METHOD_TO_SERVERLESS_CONFIG[method]
        logger.warning("Accepting End-User License Agreement by using SMTJ Serverless")
        config: Dict[str, Any] = {
            "BaseModelArn": base_model_arn,
            "AcceptEula": True,
            "JobType": "FineTuning",
            "CustomizationTechnique": technique,
        }
        if peft:
            config["Peft"] = peft
        return config

    def execute(self, job_config: JobConfig) -> str:
        from amzn_nova_forge.validation.validator import Validator

        Validator.validate_job_name(job_name=job_config.job_name)

        try:
            assert job_config.output_s3_path is not None
            assert job_config.method is not None

            with open(job_config.recipe_path) as f:
                recipe = yaml.safe_load(f)

            model_type = recipe["run"]["model_type"]
            model = Model.from_model_type(model_type)

            # Resolve base model ARN from SageMaker Hub
            base_model_arn = self._resolve_base_model_arn(model)

            # Build create_training_job request
            create_params: Dict[str, Any] = {
                "TrainingJobName": job_config.job_name,
                "RoleArn": self.execution_role,
                "OutputDataConfig": {
                    "S3OutputPath": job_config.output_s3_path,
                    "CompressionType": "NONE",
                },
                "StoppingCondition": {
                    "MaxRuntimeInSeconds": self.max_job_runtime,
                },
                "HyperParameters": self._extract_hyperparameters(recipe),
                "ServerlessJobConfig": self._build_serverless_job_config(
                    job_config.method, base_model_arn
                ),
                "ModelPackageConfig": {
                    "ModelPackageGroupArn": self.model_package_group_arn
                },
            }

            if self.kms_key_id:
                create_params["OutputDataConfig"]["KmsKeyId"] = self.kms_key_id

            # Input data via DataSet
            if job_config.data_s3_path:
                input_dataset = DataSet.create(
                    f"{job_config.job_name}-train-input", job_config.data_s3_path
                )
                create_params["InputDataConfig"] = [
                    {
                        "ChannelName": "train",
                        "DataSource": {
                            "DatasetSource": {"DatasetArn": input_dataset.arn},
                        },
                        "CompressionType": "None",
                        "RecordWrapperType": "None",
                    }
                ]
            if self.subnets or self.security_group_ids:
                create_params["VpcConfig"] = {
                    "SecurityGroupIds": self.security_group_ids or [],
                    "Subnets": self.subnets or [],
                }
            # MLflow config from recipe
            run_section = recipe.get("run", {})
            mlflow_uri = run_section.get("mlflow_tracking_uri")
            if mlflow_uri:
                create_params["MlflowConfig"] = {
                    "MlflowResourceArn": mlflow_uri,
                    "MlflowExperimentName": run_section.get(
                        "mlflow_experiment_name", ""
                    ),
                    "MlflowRunName": run_section.get("mlflow_run_name", ""),
                }

            result = self.sagemaker_client.create_training_job(**create_params)
            job_arn = result["TrainingJobArn"]
            return job_arn.split("/")[-1]

        except Exception as e:
            logger.error(f"Failed to start training job: {str(e)}")
            raise

    def cleanup(self, job_name: str) -> None:
        try:
            self.sagemaker_client.stop_training_job(TrainingJobName=job_name)
            self.sagemaker_client.close()
        except Exception as e:
            logger.error(f"Failed to cleanup job {job_name}: {str(e)}")
            raise



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validate operations and the ValidateMethod registry."""

from enum import Enum
from typing import Any, Optional

from ...util.iterator_utils import peek
from ...util.logging import logger
from .base import NovaForgeValidateOperation


class SchemaValidateOperation(NovaForgeValidateOperation):
    """Validate the dataset against schema requirements for a training method and model."""

    def execute(self, loader: Any, **kwargs) -> None:
        from ...model.model_enums import Model, TrainingMethod
        from ...recipe.recipe_config import EvaluationTask
        from ..dataset_validator import (
            CPTDatasetValidator,
            EvalDatasetValidator,
            RFTDatasetValidator,
            RFTMultiturnDatasetValidator,
            SFTDatasetValidator,
        )

        training_method: Optional[TrainingMethod] = kwargs.get("training_method")
        model: Optional[Model] = kwargs.get("model")
        eval_task: Optional[EvaluationTask] = kwargs.get("eval_task")

        if training_method is None or model is None:
            raise ValueError(
                "training_method and model are required for schema validation."
            )

        dataset_iter = loader.dataset()
        peeked_value, dataset_iter = peek(dataset_iter)

        if peeked_value is None:
            logger.info("Dataset is empty. Call load() method to load data first")
            return

        if training_method == TrainingMethod.EVALUATION and eval_task is None:
            logger.warning(
                "`eval_task` not provided for EVALUATION method. "
                "Using default evaluation validation. "
                "For RFT Multiturn evaluation, pass eval_task=EvaluationTask.RFT_MULTITURN_EVAL"
            )

        if training_method in (TrainingMethod.SFT_LORA, TrainingMethod.SFT_FULL):
            SFTDatasetValidator().validate(dataset_iter, model)
        elif training_method == TrainingMethod.EVALUATION:
            if eval_task == EvaluationTask.RFT_MULTITURN_EVAL:
                RFTMultiturnDatasetValidator(model).validate(dataset_iter, model)
            else:
                EvalDatasetValidator(eval_task).validate(dataset_iter, model)
        elif training_method in (TrainingMethod.RFT_FULL, TrainingMethod.RFT_LORA):
            RFTDatasetValidator(model).validate(dataset_iter, model)
        elif training_method in (
            TrainingMethod.RFT_MULTITURN_FULL,
            TrainingMethod.RFT_MULTITURN_LORA,
        ):
            RFTMultiturnDatasetValidator(model).validate(dataset_iter, model)
        elif training_method == TrainingMethod.CPT:
            CPTDatasetValidator().validate(dataset_iter, model)
        else:
            logger.info(
                "Skipping validation. Validation isn't available for that model/method combo right now."
            )


class ValidateMethod(Enum):
    """Supported dataset validation methods."""

    SCHEMA = "schema"


def get_validate_operation(method: ValidateMethod) -> NovaForgeValidateOperation:
    """Factory that returns the operation instance for a given ValidateMethod."""
    registry = {
        ValidateMethod.SCHEMA: SchemaValidateOperation,
    }
    op_class = registry.get(method)
    if op_class is None:
        raise ValueError(
            f"Validate method '{method.value}' is not yet implemented. "
            f"Supported: {[m.value for m in ValidateMethod]}."
        )
    return op_class()



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EC2 platform implementation for RFT Multiturn infrastructure."""

import json
import os
import time
from typing import Dict, List, Optional

import boto3

from amzn_nova_forge.util.logging import logger
from amzn_nova_forge.validation.rft_multiturn_validator import (
    validate_amazon_linux_ami,
    validate_ec2_instance_identifier,
)

from .base_infra import (
    BaseRFTInfrastructure,
    EnvType,
    StackOutputs,
)
from .common_infra_commands import BASE_PYTHON_COMMAND, CommonInfraCommands
from .constants import (
    JOB_STATUS_KILLED,
    JOB_STATUS_RUNNING,
    RFT_SAM_LOG,
    SAM_WAIT_TIME,
    SDK_RFT_LOGS_DIR,
    SDK_RFT_SCRIPTS_DIR,
    SSM_COMMAND_MAX_POLL_ATTEMPTS,
    SSM_COMMAND_POLL_INTERVAL,
    STARTER_KIT_S3,
)
from .utils import build_duplicate_job_error_message

STARTER_KIT_PATH_EC2 = "/home/ec2-user/v1"
EC2_BASE_PATH = "/home/ec2-user"
EC2_LOGS_PATH = f"{EC2_BASE_PATH}/{SDK_RFT_LOGS_DIR}"
EC2_SCRIPTS_PATH = f"{EC2_BASE_PATH}/{SDK_RFT_SCRIPTS_DIR}"


class EC2RFTInfrastructure(CommonInfraCommands, BaseRFTInfrastructure):
    """
    EC2 platform implementation
    """

    def __init__(
        self,
        region: str,
        stack_name: str,
        instance_arn: str,
        python_venv_name: str,
        rft_role_name: str,
        custom_policy_path: Optional[str] = None,
        starter_kit_path: Optional[str] = None,
        session_id: Optional[str] = None,
    ):
        super().__init__(
            region,
            stack_name,
            rft_role_name,
            custom_policy_path=custom_policy_path,
            starter_kit_path=starter_kit_path,
        )
        self.instance_id = self._extract_instance_id(instance_arn)
        self.python_venv_name = python_venv_name
        self.session_id = session_id or "default"
        self.ec2_client = boto3.client("ec2", region_name=region)
        self.ssm_client = boto3.client("ssm", region_name=region)

        self.base_path = STARTER_KIT_PATH_EC2
        self.starter_kit_s3 = STARTER_KIT_S3
        self.sam_base_dir = EC2_BASE_PATH

        # If starter_kit_path is an S3 URI, use it immediately
        if self.starter_kit_path and self.starter_kit_path.startswith("s3://"):
            self.starter_kit_s3 = self.starter_kit_path
            logger.info(f"Using custom starter kit from S3: {self.starter_kit_s3}")

        # Validate Amazon Linux AMI early during initialization
        validate_amazon_linux_ami(self.ec2_client, self.instance_id)

    def _get_base_logs_dir(self) -> str:
        """Get base logs directory path for EC2"""
        return EC2_LOGS_PATH

    def _get_script_file_path(self, session_id: str, env_type: EnvType) -> str:
        """Get script file path for EC2"""
        return f"{EC2_SCRIPTS_PATH}/{session_id}_{env_type.value}.sh"

    def _create_pid_file(self, session_id: str, env_type: EnvType, pid: int):
        """Create PID file on EC2 via SSM"""
        pid_file = self._get_pid_file_path(session_id, env_type)
        cmd = f"mkdir -p {EC2_LOGS_PATH} && echo {pid} > {pid_file}"
        self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": [cmd]},
        )

    def _read_pid_file(self, session_id: str, env_type: EnvType) -> Optional[int]:
        """Read PID from PID file on EC2 via SSM"""
        pid_file = self._get_pid_file_path(session_id, env_type)
        cmd = f"cat {pid_file} 2>/dev/null || echo ''"

        response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": [cmd]},
        )

        # Wait for command to complete
        command_id = response["Command"]["CommandId"]
        try:
            self._wait_for_ssm_command(command_id, timeout=10)
            output = self.ssm_client.get_command_invocation(
                CommandId=command_id, InstanceId=self.instance_id
            )
            pid_str = output["StandardOutputContent"].strip()
            return int(pid_str) if pid_str else None
        except Exception:
            return None

    def _create_status_file(self, session_id: str, env_type: EnvType, status: str):
        """Create or update status file on EC2 via SSM"""
        status_file = self._get_status_file_path(session_id, env_type)
        cmd = f"mkdir -p {EC2_LOGS_PATH} && echo {status} > {status_file}"
        self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": [cmd]},
        )

    def _read_status_file(self, session_id: str, env_type: EnvType) -> Optional[str]:
        """Read status from status file on EC2 via SSM"""
        status_file = self._get_status_file_path(session_id, env_type)
        cmd = f"cat {status_file} 2>/dev/null || echo ''"

        response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": [cmd]},
        )

        # Wait for command to complete
        command_id = response["Command"]["CommandId"]
        try:
            self._wait_for_ssm_command(command_id, timeout=10)
            output = self.ssm_client.get_command_invocation(
                CommandId=command_id, InstanceId=self.instance_id
            )
            status = output["StandardOutputContent"].strip()
            return status if status else None
        except Exception:
            return None

    def _check_for_running_jobs_on_stack(self, stack_name: str) -> Optional[str]:
        """
        Check if any jobs are running for this stack on EC2 (across all sessions).

        Args:
            stack_name: Stack name to check

        Returns:
            Error message if running jobs found, None otherwise
        """
        base_stack = self._extract_base_stack_name(stack_name)

        check_cmd = (
            f"echo 'TRAIN:'; "
            f"(pgrep -f '{stack_name}_.*_train\\.sh' 2>/dev/null || true; "
            f"pgrep -f '{base_stack}_.*_train\\.sh' 2>/dev/null || true) | sort -u || echo 'none'; "
            f"echo 'EVAL:'; "
            f"(pgrep -f '{stack_name}_.*_eval\\.sh' 2>/dev/null || true; "
            f"pgrep -f '{base_stack}_.*_eval\\.sh' 2>/dev/null || true) | sort -u || echo 'none'"
        )

        try:
            response = self.ssm_client.send_command(
                InstanceIds=[self.instance_id],
                DocumentName="AWS-RunShellScript",
                Parameters={"commands": [check_cmd]},
            )

            command_id = response["Command"]["CommandId"]

            for attempt in range(SSM_COMMAND_MAX_POLL_ATTEMPTS):
                try:
                    result = self.ssm_client.get_command_invocation(
                        CommandId=command_id, InstanceId=self.instance_id
                    )
                    status = result.get("Status")

                    if status in ["Success", "Failed"]:
                        stdout = result.get("StandardOutputContent", "").strip()

                        train_pids = []
                        eval_pids = []

                        lines = stdout.split("\n")
                        current_type = None
                        for line in lines:
                            line = line.strip()
                            if line == "TRAIN:":
                                current_type = "train"
                            elif line == "EVAL:":
                                current_type = "eval"
                            elif line and line != "none":
                                if current_type == "train":
                                    train_pids.append(line)
                                elif current_type == "eval":
                                    eval_pids.append(line)

                        if train_pids or eval_pids:
                            train_jobs = (
                                [f"PIDs: {', '.join(train_pids)}"] if train_pids else []
                            )
                            eval_jobs = (
                                [f"PIDs: {', '.join(eval_pids)}"] if eval_pids else []
                            )

                            error_msg = build_duplicate_job_error_message(
                                stack_name,
                                train_jobs,
                                eval_jobs,
                                f"on EC2 instance {self.instance_id}",
                            )
                            logger.warning(error_msg)
                            return error_msg

                        return None
                    elif status in ["Pending", "InProgress"]:
                        time.sleep(SSM_COMMAND_POLL_INTERVAL)
                except Exception as e:
                    if "InvocationDoesNotExist" in str(e):
                        time.sleep(SSM_COMMAND_POLL_INTERVAL)
                    else:
                        raise

            return None

        except Exception as e:
            logger.warning(f"Could not check for running jobs: {e}")
            return None

    @staticmethod
    def _extract_instance_id(arn: str) -> str:
        """
        Extract instance ID from ARN or return as-is if already ID
        """
        return validate_ec2_instance_identifier(arn)

    def _get_package_install_cmd(self) -> List[str]:
        return [
            "sudo yum update -y",
            f"sudo yum install -y {BASE_PYTHON_COMMAND} {BASE_PYTHON_COMMAND}-pip git",
        ]

    def _wait_for_ssm_command(self, command_id: str, timeout: int = 300) -> None:
        """
        Wait for SSM command to complete successfully
        """
        logger.info(f"Waiting for SSM command {command_id} to complete...")
        wait_interval = 10
        elapsed = 0

        while elapsed < timeout:
            time.sleep(wait_interval)
            elapsed += wait_interval

            try:
                invocation = self.ssm_client.get_command_invocation(
                    CommandId=command_id, InstanceId=self.instance_id
                )
                status = invocation["Status"]

                if status == "Success":
                    logger.info("Environment setup completed successfully")
                    return
                elif status in ["Failed", "Cancelled", "TimedOut"]:
                    raise RuntimeError(f"SSM command failed with status: {status}")
            except self.ssm_client.exceptions.InvocationDoesNotExist:
                continue

        logger.warning(f"SSM command still running after {timeout}s, proceeding anyway")

    def _execute_training_or_eval(
        self, vf_env_id: str, script_name: str, log_file: str, command_builder
    ):
        """
        Execute training or evaluation command on EC2 instance via SSM

        Checks for duplicate processes before starting.
        Creates folders and tracking files.
        """
        # Check if ANY job is running for this stack (across all sessions)
        error_msg = self._check_for_running_jobs_on_stack(self.stack_name)
        if error_msg:
            raise RuntimeError(error_msg)

        logger.info(
            f"Setting up environment for '{vf_env_id}' on EC2 instance {self.instance_id}"
        )
        install_commands = self._build_setup_commands(vf_env_id)
        setup_response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": install_commands},
        )

        self._wait_for_ssm_command(setup_response["Command"]["CommandId"])

        cmd = command_builder()
        script_content = f"#!/bin/bash\n{cmd}"

        # Create folders and write script
        commands = [
            f"mkdir -p {EC2_LOGS_PATH}",
            f"mkdir -p {EC2_SCRIPTS_PATH}",
            f"cat > {EC2_SCRIPTS_PATH}/{script_name} << 'EOFSCRIPT'\n{script_content}\nEOFSCRIPT",
            f"chmod +x {EC2_SCRIPTS_PATH}/{script_name}",
            f"nohup {EC2_SCRIPTS_PATH}/{script_name} > {EC2_LOGS_PATH}/{log_file} 2>&1 </dev/null & echo $!",
        ]

        response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": commands},
        )

        # Get the PID from the command output
        command_id = response["Command"]["CommandId"]
        self._wait_for_ssm_command(command_id, timeout=30)

        try:
            output = self.ssm_client.get_command_invocation(
                CommandId=command_id, InstanceId=self.instance_id
            )
            pid_str = output["StandardOutputContent"].strip().split("\n")[-1]
            pid = int(pid_str)

            # Determine env_type from script_name
            env_type = EnvType.TRAIN if "train" in script_name else EnvType.EVAL

            # Create tracking files
            self._create_pid_file(self.session_id, env_type, pid)
            self._create_status_file(self.session_id, env_type, JOB_STATUS_RUNNING)
            logger.info(
                f"Created tracking files for session {self.session_id}, PID: {pid}"
            )
        except Exception as e:
            logger.warning(f"Could not create tracking files: {e}")

    def validate_platform(self):
        """
        Validate EC2 instance exists, is ready, and has required IAM role with RFT permissions
        """
        response = self.ec2_client.describe_instances(InstanceIds=[self.instance_id])
        if not response["Reservations"]:
            raise ValueError(f"Instance {self.instance_id} not found")

        instance = response["Reservations"][0]["Instances"][0]
        if instance["State"]["Name"] != "running":
            raise ValueError(f"Instance {self.instance_id} is not running")

        # Check SSM connectivity
        self._validate_ssm_connectivity()

        # Check if instance has an IAM instance profile
        if "IamInstanceProfile" not in instance:
            raise ValueError(
                f"Instance {self.instance_id} does not have an IAM instance profile attached. "
                f"Please attach an instance profile with an IAM role."
            )

        # Get the instance profile and ensure it has RFT permissions
        iam_client = boto3.client("iam", region_name=self.region)
        sts_client = boto3.client("sts")
        account_id = sts_client.get_caller_identity()["Account"]

        instance_profile_arn = instance["IamInstanceProfile"]["Arn"]
        instance_profile_name = instance_profile_arn.split("/")[-1]

        try:
            profile = iam_client.get_instance_profile(
                InstanceProfileName=instance_profile_name
            )
            roles = profile["InstanceProfile"]["Roles"]

            if not roles:
                raise ValueError(
                    f"Instance profile {instance_profile_name} has no roles attached. "
                    f"Please attach an IAM role to the instance profile."
                )

            # Get the role and ensure it has RFT permissions
            role_name = roles[0]["RoleName"]
            logger.info(f"EC2 instance using IAM role: {role_name}")

            # Attach RFT policy to the EC2 instance role
            self._ensure_policy_on_role(role_name)

        except Exception as e:
            raise ValueError(
                f"Failed to validate or configure IAM role permissions: {e}"
            )

    def _validate_ssm_connectivity(self):
        """
        Validate that SSM agent is connected and can receive commands
        """
        try:
            response = self.ssm_client.describe_instance_information(
                Filters=[{"Key": "InstanceIds", "Values": [self.instance_id]}]
            )

            if not response["InstanceInformationList"]:
                raise ValueError(
                    f"Instance {self.instance_id} is not registered with SSM. "
                    f"Please ensure SSM agent is installed and running."
                )

            instance_info = response["InstanceInformationList"][0]
            ping_status = instance_info.get("PingStatus")

            if ping_status != "Online":
                raise ValueError(
                    f"Instance {self.instance_id} SSM agent status is '{ping_status}'. "
                    f"Expected 'Online'. Please restart the SSM agent or reboot the instance."
                )

            logger.info(f"SSM connectivity validated for instance {self.instance_id}")

        except Exception as e:
            if "is not registered with SSM" in str(e) or "SSM agent status" in str(e):
                raise
            raise ValueError(
                f"Failed to validate SSM connectivity for instance {self.instance_id}: {e}"
            )

    def _ensure_policy_on_role(self, role_name: str):
        """
        Ensure RFT policy is attached to the specified role
        """
        self.attach_rft_policy_to_role(role_name)

    def _ensure_starter_kit_available(self, s3_bucket: Optional[str] = None):
        """
        Ensure starter kit is available for remote deployment.

        For EC2, this just handles S3 upload. Setup commands are run by:
        - deploy_sam_stack() via _build_sam_deploy_commands()
        - start_environment() via _build_setup_commands()

        Args:
            s3_bucket: S3 bucket for upload (optional, uses SageMaker default if not provided)
        """
        # Just use base implementation - no EC2-specific setup needed here
        super()._ensure_starter_kit_available(s3_bucket)

    def deploy_sam_stack(self, s3_bucket: Optional[str] = None):
        """
        Deploy SAM via EC2 instance

        Args:
            s3_bucket: S3 bucket for starter kit upload (if needed)
        """
        # Ensure starter kit is available
        self._ensure_starter_kit_available(s3_bucket)

        # Set log file for SAM deployment
        self.sam_log_file = f"{EC2_BASE_PATH}/{RFT_SAM_LOG}"

        commands = self._build_sam_deploy_commands(
            custom_starter_kit_s3=self._starter_kit_s3_uri
        )

        response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": commands},
        )
        command_id = response["Command"]["CommandId"]
        logger.info(f"SAM deployment initiated on EC2 instance {self.instance_id}")

        # Wait for command to complete
        start_time = time.time()

        while time.time() - start_time < SAM_WAIT_TIME:
            try:
                result = self.ssm_client.get_command_invocation(
                    CommandId=command_id, InstanceId=self.instance_id
                )
                status = result["Status"]
            except self.ssm_client.exceptions.InvocationDoesNotExist:
                continue

            if status == "Success":
                logger.info("SAM deployment completed successfully")
                return
            elif status in ["Failed", "Cancelled", "TimedOut"]:
                error_msg = result.get("StandardErrorContent", "Unknown error")
                raise RuntimeError(f"SAM deployment failed: {error_msg}")

        raise TimeoutError("SAM deployment timed out after 10 minutes")

    def start_environment(
        self,
        env_type: EnvType,
        vf_env_id: str,
        vf_env_args: Dict,
        stack_outputs: StackOutputs,
        max_concurrent_rollouts: int = 40,
        max_rollout_timeout: float = 300.0,
        completion_poll_timeout: float = 600.0,
        completion_poll_interval: float = 0.5,
        rollout_poll_interval: float = 1.0,
        log_output_directory: Optional[str] = None,
        config_name: Optional[str] = None,
        config_path: Optional[str] = None,
    ):
        """
        Start environment on EC2 using unified environment client.

        Args:
            env_type: Environment type (EnvType.TRAIN or EnvType.EVAL) for script and log file naming
            vf_env_id: Verifier environment identifier
            vf_env_args: Environment arguments
            stack_outputs: Stack outputs containing URLs and queue information
            max_concurrent_rollouts: Max concurrent rollouts (default: 40)
            max_rollout_timeout: Per-rollout timeout in seconds (default: 300)
            completion_poll_timeout: Completion polling timeout (default: 600)
            completion_poll_interval: Completion poll interval (default: 0.5)
            rollout_poll_interval: SQS polling interval (default: 1.0)
            log_output_directory: Directory for logs and metrics (optional)
            config_name: Use YAML config instead of CLI flags (optional)
            config_path: Custom config directory path (optional, use with config_name)

        Note:
            Script names include session ID to allow multiple sessions/stacks on the same EC2 instance.
            Each session can be managed independently.
        """
        logger.warning(
            f"Starting {env_type.value} environment on EC2. "
            f"Note: Checking for duplicate processes in this session. "
            f"Multiple stacks/sessions can run simultaneously on the same instance."
        )

        def command_builder():
            return self._build_unified_client_command(
                vf_env_id=vf_env_id,
                vf_env_args=vf_env_args,
                lambda_url=stack_outputs.proxy_function_url,
                queue_url=stack_outputs.rollout_request_queue_url,
                max_concurrent_rollouts=max_concurrent_rollouts,
                max_rollout_timeout=max_rollout_timeout,
                completion_poll_timeout=completion_poll_timeout,
                completion_poll_interval=completion_poll_interval,
                rollout_poll_interval=rollout_poll_interval,
                log_output_directory=log_output_directory,
                config_name=config_name,
                config_path=config_path,
            )

        # Use new naming format: {session_id}_{env_type}.sh and {session_id}_{env_type}.log
        session_id = getattr(self, "session_id", "default")
        script_name = f"{session_id}_{env_type.value}.sh"
        log_file = f"{session_id}_{env_type.value}.log"

        self._execute_training_or_eval(
            vf_env_id, script_name, log_file, command_builder
        )

        logger.info(f"Environment started on EC2 instance {self.instance_id}")

    def get_logs(
        self,
        env_type: EnvType,
        limit: int,
        start_from_head: bool,
        log_stream_name: Optional[str],
        tail: bool = False,
    ) -> list:
        """
        Get logs from EC2 via SSM for this session
        """
        # Use session-specific log file in new folder structure
        session_id = getattr(self, "session_id", "default")
        log_file = self._get_log_file_path(session_id, env_type)

        if tail:
            logger.info(
                f"Tailing {log_file} from EC2 instance {self.instance_id} (Press Ctrl+C to stop)"
            )
            try:
                last_lines = set()
                while True:
                    response = self.ssm_client.send_command(
                        InstanceIds=[self.instance_id],
                        DocumentName="AWS-RunShellScript",
                        Parameters={
                            "commands": [
                                f"tail -n 50 {log_file} 2>/dev/null || echo 'Log file not found'"
                            ]
                        },
                    )

                    # Wait for command to complete
                    command_id = response["Command"]["CommandId"]
                    self._wait_for_ssm_command(command_id)

                    output = self.ssm_client.get_command_invocation(
                        CommandId=command_id,
                        InstanceId=self.instance_id,
                    )
                    if output["StandardOutputContent"]:
                        lines = output["StandardOutputContent"].strip().split("\n")
                        for line in lines:
                            if line not in last_lines:
                                logger.info(line)
                                last_lines.add(line)
            except KeyboardInterrupt:
                logger.info("Stopped tailing logs")
            return []

        response = self.ssm_client.send_command(
            InstanceIds=[self.instance_id],
            DocumentName="AWS-RunShellScript",
            Parameters={
                "commands": [
                    f"tail -n {limit} {log_file} 2>/dev/null || echo 'Log file not found'"
                ]
            },
        )

        # Wait for command to complete
        command_id = response["Command"]["CommandId"]
        self._wait_for_ssm_command(command_id)

        output = self.ssm_client.get_command_invocation(
            CommandId=command_id, InstanceId=self.instance_id
        )
        return (
            output["StandardOutputContent"].strip().split("\n")
            if output["StandardOutputContent"]
            else []
        )

    def kill_task(
        self,
        env_type: EnvType,
        kill_all_for_stack: bool = False,
        preserve_logs: bool = True,
    ):
        """
        Kill training or evaluation task on EC2.

        Args:
            env_type: Type of environment (TRAIN or EVAL)
            kill_all_for_stack: If True, kills ALL jobs of this type for the stack (cross-session).
                               If False, only kills jobs from current session.
            preserve_logs: If True, keeps log files after killing jobs. If False, deletes them.
        """
        if kill_all_for_stack:
            base_stack = self._extract_base_stack_name(self.stack_name)

            logger.info(
                f"Killing ALL {env_type.value} jobs for stack '{self.stack_name}' on EC2"
            )

            if preserve_logs:
                cmd = f"""
                PIDS=$({{ pgrep -f '{self.stack_name}_.*_{env_type.value}\\.sh' 2>/dev/null || true; pgrep -f '{base_stack}_.*_{env_type.value}\\.sh' 2>/dev/null || true; }} | sort -u)
                
                if [ -n "$PIDS" ]; then
                    echo "Found PIDs: $PIDS"
                    pkill -f '{self.stack_name}_.*_{env_type.value}\\.sh' 2>/dev/null || true
                    pkill -f '{base_stack}_.*_{env_type.value}\\.sh' 2>/dev/null || true
                    
                    cd {EC2_LOGS_PATH} 2>/dev/null || exit 0
                    for file in {self.stack_name}_*_{env_type.value}.* {base_stack}_*_{env_type.value}.*; do
                        if [ -f "$file" ]; then
                            SESSION_ID=$(echo "$file" | sed 's/_{env_type.value}\\..*$//')
                            echo "{JOB_STATUS_KILLED}" > "${{SESSION_ID}}_{env_type.value}.status" 2>/dev/null || true
                        fi
                    done
                    
                    cd {EC2_SCRIPTS_PATH} 2>/dev/null || exit 0
                    rm -f {self.stack_name}_*_{env_type.value}.sh 2>/dev/null || true
                    rm -f {base_stack}_*_{env_type.value}.sh 2>/dev/null || true
                    
                    echo "Killed all {env_type.value} jobs (logs preserved)"
                else
                    echo "No {env_type.value} jobs found"
                fi
                """
            else:
                cmd = f"""
                PIDS=$({{ pgrep -f '{self.stack_name}_.*_{env_type.value}\\.sh' 2>/dev/null || true; pgrep -f '{base_stack}_.*_{env_type.value}\\.sh' 2>/dev/null || true; }} | sort -u)
                
                if [ -n "$PIDS" ]; then
                    echo "Found PIDs: $PIDS"
                    pkill -f '{self.stack_name}_.*_{env_type.value}\\.sh' 2>/dev/null || true
                    pkill -f '{base_stack}_.*_{env_type.value}\\.sh' 2>/dev/null || true
                    
                    cd {EC2_LOGS_PATH} 2>/dev/null || exit 0
                    for file in {self.stack_name}_*_{env_type.value}.* {base_stack}_*_{env_type.value}.*; do
                        if [ -f "$file" ]; then
                            SESSION_ID=$(echo "$file" | sed 's/_{env_type.value}\\..*$//')
                            echo "{JOB_STATUS_KILLED}" > "${{SESSION_ID}}_{env_type.value}.status" 2>/dev/null || true
                            rm -f "${{SESSION_ID}}_{env_type.value}.log" 2>/dev/null || true
                            rm -f "${{SESSION_ID}}_{env_type.value}.pid" 2>/dev/null || true
                            rm -f "${{SESSION_ID}}_{env_type.value}.status" 2>/dev/null || true
                        fi
                    done
                    
                    cd {EC2_SCRIPTS_PATH} 2>/dev/null || exit 0
                    rm -f {self.stack_name}_*_{env_type.value}.sh 2>/dev/null || true
                    rm -f {base_stack}_*_{env_type.value}.sh 2>/dev/null || true
                    
                    echo "Killed all {env_type.value} jobs"
                else
                    echo "No {env_type.value} jobs found"
                fi
                """

            response = self.ssm_client.send_command(
                InstanceIds=[self.instance_id],
                DocumentName="AWS-RunShellScript",
                Parameters={"commands": [cmd]},
            )

            command_id = response["Command"]["CommandId"]
            self._wait_for_ssm_command(command_id)

            output = self.ssm_client.get_command_invocation(
                CommandId=command_id,
                InstanceId=self.instance_id,
            )

            if output.get("StandardOutputContent"):
                logger.info(f"Kill output: {output['StandardOutputContent'].strip()}")

            logger.info(f"All {env_type.value} jobs killed on EC2")
        else:
            session_id = getattr(self, "session_id", "default")
            script_path = self._get_script_file_path(session_id, env_type)
            log_file = self._get_log_file_path(session_id, env_type)
            pid_file = self._get_pid_file_path(session_id, env_type)
            status_file = self._get_status_file_path(session_id, env_type)

            script_name = f"{session_id}_{env_type.value}.sh"

            if preserve_logs:
                cmd = f"pkill -f {script_name}; echo {JOB_STATUS_KILLED} > {status_file}; rm -f {script_path} {pid_file}"
            else:
                cmd = f"pkill -f {script_name}; echo {JOB_STATUS_KILLED} > {status_file}; rm -f {script_path} {log_file} {pid_file} {status_file}"

            self.ssm_client.send_command(
                InstanceIds=[self.instance_id],
                DocumentName="AWS-RunShellScript",
                Parameters={"commands": [cmd]},
            )
            logger.info(f"{env_type.value} task killed on EC2")

    def cleanup(self, cleanup_environment: bool = False):
        """
        Clean up EC2 resources: processes, logs, and optionally environment

        Args:
            cleanup_environment: If True, delete virtual environment and starter kit directories
        """
        self.kill_task(EnvType.TRAIN)
        self.kill_task(EnvType.EVAL)

        if cleanup_environment:
            # Remove virtual environment and starter kit
            commands = [
                f"rm -rf {STARTER_KIT_PATH_EC2}/{self.python_venv_name}",
                f"rm -rf {STARTER_KIT_PATH_EC2}",
                f"rm -rf {EC2_LOGS_PATH}",
                f"rm -rf {EC2_SCRIPTS_PATH}",
                f"rm -f {EC2_BASE_PATH}/{RFT_SAM_LOG}",
            ]

            # Also remove custom environment if it was used
            if (
                hasattr(self, "custom_env")
                and self.custom_env
                and self.custom_env.s3_uri
            ):
                vf_env_id = self.custom_env.env_id
                parent_path = EC2_BASE_PATH.rsplit("/", 1)[0]
                custom_env_dir = f"{parent_path}/{vf_env_id}"
                custom_env_tarball = f"/tmp/{vf_env_id}.tar.gz"
                commands.extend(
                    [
                        f"rm -rf {custom_env_dir}",
                        f"rm -f {custom_env_tarball}",
                    ]
                )

            self.ssm_client.send_command(
                InstanceIds=[self.instance_id],
                DocumentName="AWS-RunShellScript",
                Parameters={"commands": commands},
            )
            logger.info(
                f"Deleted environment and logs on EC2 instance {self.instance_id}"
            )

        logger.info(f"EC2 cleanup complete for instance {self.instance_id}")

    def get_state(self) -> Dict:
        """Get EC2 platform state for serialization"""
        return {
            "instance_id": self.instance_id,
            "python_venv_name": self.python_venv_name,
            "base_path": self.base_path,
            "starter_kit_s3": self.starter_kit_s3,
        }

    def restore_state(self, state: Dict):
        """Restore EC2 platform state after deserialization"""
        # Restore custom starter kit S3 URI if it was set
        if "starter_kit_s3" in state:
            self.starter_kit_s3 = state["starter_kit_s3"]
        # EC2 doesn't need special restoration - instance_id is already set in __init__
        # Verify instance is still running
        try:
            response = self.ec2_client.describe_instances(
                InstanceIds=[self.instance_id]
            )
            if response["Reservations"]:
                status = response["Reservations"][0]["Instances"][0]["State"]["Name"]
                logger.info(f"EC2 instance {self.instance_id} status: {status}")
        except Exception as e:
            logger.warning(f"Could not verify EC2 instance: {e}")



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Operation for splitting datasets into train/validation/test sets."""

import random
from typing import Any, Dict, Iterator, List, Optional, Tuple

from ...util.logging import logger
from .base import DataPrepError, NovaForgeSplitOperation


class SplitOperation(NovaForgeSplitOperation):
    """Split the dataset into train, validation, and test DatasetLoader objects."""

    def execute(self, loader: Any, **kwargs) -> Tuple[Any, Any, Any]:
        """
        Split the current dataset into train/val/test loaders.

        Args:
            loader: The DatasetLoader instance.
            train_ratio: Fraction for training.
            val_ratio: Fraction for validation.
            test_ratio: Fraction for testing.
            seed: Random seed.

        Returns:
            Tuple of (train_loader, val_loader, test_loader).
        """
        train_ratio, val_ratio, test_ratio = self._resolve_ratios(
            kwargs.get("train_ratio"),
            kwargs.get("val_ratio"),
            kwargs.get("test_ratio"),
        )
        seed: int = kwargs.get("seed", 42)

        dataset = self._materialize_dataset(loader)
        indices = self._shuffled_indices(len(dataset), seed)
        n_train, n_val, n_test = self._compute_split_sizes(
            len(dataset), train_ratio, val_ratio, test_ratio
        )

        train_indices = indices[:n_train]
        val_indices = indices[n_train : n_train + n_val]
        test_indices = indices[n_train + n_val :]

        train_loader = self._create_split_loader(loader, dataset, train_indices)
        val_loader = self._create_split_loader(loader, dataset, val_indices)
        test_loader = self._create_split_loader(loader, dataset, test_indices)

        logger.info(f"Data split: {n_train} train, {n_val} val, {n_test} test")
        return train_loader, val_loader, test_loader

    # ------------------------------------------------------------------
    # Private helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _resolve_ratios(
        train_ratio: Optional[float],
        val_ratio: Optional[float],
        test_ratio: Optional[float],
    ) -> Tuple[float, float, float]:
        """Return validated (train, val, test) ratios, applying defaults when all are None."""
        if (train_ratio, val_ratio, test_ratio) == (None, None, None):
            return 0.8, 0.1, 0.1

        if train_ratio is None or val_ratio is None or test_ratio is None:
            raise DataPrepError(
                f"Please provide three values for split_data: train_ratio, val_ratio, and test_ratio."
                f"You provided: (Train: {train_ratio}, Val: {val_ratio}, Test: {test_ratio})"
            )

        if any(r < 0 for r in [train_ratio, val_ratio, test_ratio]):
            raise DataPrepError(
                "Calculated ratio is negative. Provided ratios sum to > 1.0"
            )

        if abs(train_ratio + val_ratio + test_ratio - 1.0) > 1e-6:
            raise DataPrepError(
                f"Split ratios must sum to 1.0. Current Ratios: "
                f"(Train: {train_ratio}, Val: {val_ratio}, Test: {test_ratio} "
                f"-> Total: {abs(train_ratio + val_ratio + test_ratio)})"
            )

        return train_ratio, val_ratio, test_ratio

    @staticmethod
    def _materialize_dataset(loader: Any) -> List[Dict]:
        """Load the full dataset into memory, raising on empty."""
        dataset = list(loader.dataset())
        if not dataset:
            raise DataPrepError("Dataset is empty. Call load() method first")

        if len(dataset) < 10:
            logger.info(
                "The provided dataset is small. Data will be split, "
                "but consider adding more data for better results."
            )
        return dataset

    @staticmethod
    def _shuffled_indices(n: int, seed: int) -> List[int]:
        """Return a shuffled list of indices [0, n)."""
        random.seed(seed)
        indices = list(range(n))
        random.shuffle(indices)
        return indices

    @staticmethod
    def _compute_split_sizes(
        n_total: int,
        train_ratio: float,
        val_ratio: float,
        test_ratio: float,
    ) -> Tuple[int, int, int]:
        """Convert ratios to concrete counts that sum to n_total."""
        n_train = max(1, round(n_total * train_ratio)) if train_ratio > 0 else 0
        remaining = n_total - n_train

        if val_ratio == 0:
            n_val = 0
            n_test = remaining
        elif test_ratio == 0:
            n_val = remaining
            n_test = 0
        else:
            n_val = max(1, round(n_total * val_ratio)) if val_ratio > 0 else 0
            n_test = remaining - n_val

        # Ensure we haven't exceeded the total length
        if n_train + n_val + n_test > n_total:
            if n_test > 0:
                n_test -= 1
            elif n_val > 0:
                n_val -= 1
            else:
                n_train -= 1

        return n_train, n_val, n_test

    @staticmethod
    def _create_split_loader(
        loader: Any, dataset: List[Dict], split_indices: List[int]
    ) -> Any:
        """Create a new loader of the same type backed by the given indices."""

        def generator() -> Iterator[Dict]:
            for idx in split_indices:
                yield dataset[idx]

        new_loader = loader.__class__(**loader.column_mappings)
        new_loader.dataset = generator
        return new_loader



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Dict, List, Optional

import pandas

from amzn_nova_forge.model.model_enums import Platform, TrainingMethod

GLOBAL_STEP_REGEX = r"global_step[=:]\s*([\d.]+)"
TRAINING_LOSS_REGEX = r"reduced_train_loss[=:]\s*(-?[\d.]+(?:[eE][+-]?\d+)?)"
SMHP_RFT_REWARD_SCORE_REGEX = r"train_rm_score:\s*(-?[\d.]+(?:[eE][+-]?\d+)?)"
SMTJ_RFT_REWARD_SCORE_REGEX = r"critic/rewards/mean[=:]\s*(-?[\d.]+(?:[eE][+-]?\d+)?)"
CPT = "cpt"
SFT = "sft"
RFT = "rft"
AVAILABLE_METRICS = {
    Platform.SMTJ: {
        CPT: {"training_loss": TRAINING_LOSS_REGEX},
        SFT: {"training_loss": TRAINING_LOSS_REGEX},
        RFT: {"reward_score": SMTJ_RFT_REWARD_SCORE_REGEX},
    },
    Platform.SMHP: {
        CPT: {"training_loss": TRAINING_LOSS_REGEX},
        SFT: {"training_loss": TRAINING_LOSS_REGEX},
        RFT: {"reward_score": SMHP_RFT_REWARD_SCORE_REGEX},
    },
}


def get_metrics(
    platform: Platform,
    training_method: TrainingMethod,
    logs: Optional[List[Dict]] = None,
    metrics: Optional[List] = None,
) -> pandas.DataFrame:
    patterns = []
    training_category = training_method.value[:3]
    if not metrics:
        metrics = list(AVAILABLE_METRICS[platform][training_category].keys())
    for metric in metrics:
        if metric not in AVAILABLE_METRICS[platform][training_category]:
            raise NotImplementedError(
                f"Unsupported metric for {training_category} on {platform}: {metric}"
            )
        patterns.append(AVAILABLE_METRICS[platform][training_category][metric])

    all_metrics: List[List[float]] = []
    log_lines = [line for log in (logs or []) for line in log["message"].splitlines()]

    for line in log_lines:
        global_step_match = re.search(GLOBAL_STEP_REGEX, line)
        if not global_step_match:
            continue
        try:
            step_metrics: List[float] = [int(float(global_step_match.group(1)))]
            for pattern in patterns:
                match = re.search(pattern, line)
                if match is None:
                    raise ValueError(f"Pattern {pattern} not found in line")
                step_metrics.append(float(match.group(1)))
            all_metrics.append(step_metrics)
        except Exception:
            pass

    return pandas.DataFrame(all_metrics, columns=["global_step"] + metrics)



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Operation for saving datasets to local or S3 storage."""

from typing import Any

from ...util.dataset_writer import DatasetWriter
from ...util.iterator_utils import peek
from ...util.logging import logger
from .base import DataPrepError, NovaForgeSaveOperation


class SaveOperation(NovaForgeSaveOperation):
    """Save the current dataset to a local file or S3 path."""

    def execute(self, loader: Any, **kwargs) -> str:
        """
        Save the dataset to the specified path.

        Args:
            loader: The DatasetLoader instance.
            save_path: Path where to save the file (.json or .jsonl).

        Returns:
            The path where the file was saved.

        Raises:
            DataPrepError: If the format is unsupported or saving fails.
        """
        save_path: str = kwargs["save_path"]

        # Get iterator from the current dataset
        dataset_iter = loader.dataset()
        peeked_value, dataset_iter = peek(dataset_iter)

        if peeked_value is None:
            logger.warning("Warning: Dataset is empty. An empty dataset will be saved.")
            dataset_iter = iter([])

        try:
            # Determine format
            if save_path.endswith(".jsonl"):
                is_jsonl = True
            elif save_path.endswith(".json"):
                is_jsonl = False
            else:
                raise DataPrepError(
                    "Unsupported format. Use '.json' or '.jsonl' extension"
                )

            # Save to S3 or local file using DatasetWriter
            if save_path.startswith("s3://"):
                DatasetWriter.save_to_s3(save_path, dataset_iter, is_jsonl)
            else:
                DatasetWriter.save_to_local(save_path, dataset_iter, is_jsonl)

            logger.info(f"Dataset saved successfully to {save_path}")
            return save_path

        except Exception as e:
            raise DataPrepError(f"Error saving dataset: {str(e)}")



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for RFT Multiturn infrastructure."""

import os
from typing import List


def validate_starter_kit_path(path: str, raise_on_invalid: bool = True) -> bool:
    """
    Validate if a path is a valid starter kit directory by checking for lambda_proxy subdirectory.

    Args:
        path: Path to validate
        raise_on_invalid: If True, raises ValueError on invalid path. If False, returns False.

    Returns:
        True if valid starter kit, False if invalid (only when raise_on_invalid=False)

    Raises:
        ValueError: If path is not a valid starter kit and raise_on_invalid=True
    """
    lambda_proxy_dir = os.path.join(path, "lambda_proxy")

    if not os.path.exists(lambda_proxy_dir):
        if raise_on_invalid:
            raise ValueError(
                f"Invalid starter kit directory (missing lambda_proxy): {path}\n"
                f"Expected structure: {path}/lambda_proxy/"
            )
        return False

    return True


def build_duplicate_job_error_message(
    stack_name: str,
    train_jobs: List[str],
    eval_jobs: List[str],
    platform_info: str = "",
) -> str:
    """
    Build standardized error message for duplicate job detection.

    Args:
        stack_name: Stack name
        train_jobs: List of train job identifiers
        eval_jobs: List of eval job identifiers
        platform_info: Optional platform-specific info (e.g., "on EC2 instance i-123")

    Returns:
        Formatted error message
    """
    job_details = []
    kill_instructions = []

    if train_jobs:
        job_details.append(f"{len(train_jobs)} TRAIN job(s) ({', '.join(train_jobs)})")
        kill_instructions.append(
            "rft_infra.kill_task(env_type=EnvType.TRAIN, kill_all_for_stack=True)"
        )
    if eval_jobs:
        job_details.append(f"{len(eval_jobs)} EVAL job(s) ({', '.join(eval_jobs)})")
        kill_instructions.append(
            "rft_infra.kill_task(env_type=EnvType.EVAL, kill_all_for_stack=True)"
        )

    platform_suffix = f" {platform_info}" if platform_info else ""

    return (
        f"Found running job(s) for stack '{stack_name}'{platform_suffix}:\n"
        f"  - {' and '.join(job_details)}\n"
        f"To stop them, run:\n"
        f"  {chr(10).join('  ' + instr for instr in kill_instructions)}"
    )
